import produce, { Draft } from 'immer';
import React, { useMemo, useRef } from 'react';
import { useReducer } from 'react';
import { createContext, useContextSelector } from 'use-context-selector';
import { shallowEqual } from './utils';

export type Action<P, T = string> = {
  type: T;
  payload: P;
};

export type Dispatch<
  T extends {
    useDispatch?: any;
  }
> = ReturnType<T['useDispatch']>;

export type Selector<T> = T extends {
  useSelector: (selector: (state: infer S) => any) => any;
}
  ? (state: S) => any
  : never;

export type Reducer<T, P> = (
  state: Draft<T>,
  action: Action<P, string>
) => void;

type PayloadParams<T> = T extends (
  a: any,
  action: {
    type: any;
    payload: infer P;
  }
) => any
  ? P
  : never;

export function createKit<T, R extends { [key: string]: Reducer<T, any> }>({
  initialState,
  reducers,
}: {
  initialState: T;
  reducers: R;
}) {
  const Context = createContext<any>(null);

  const Provider: React.FC<{ children?: React.ReactNode }> = ({ children }) => {
    function reducerWithImmer<T>(
      reducers: Record<string, Reducer<T, any>> = {}
    ) {
      return function combination(state: T, action: Action<any, any>) {
        const next = produce(state, (draft) => {
          const key = action.type;
          reducers[key](draft, action);
        });
        return next;
      };
    }

    const [state, dispatch] = useReducer(
      reducerWithImmer(reducers),
      initialState
    );
    const snapshots = useRef<any[]>([]);

    return (
      <Context.Provider value={[state, dispatch, snapshots]}>
        {children}
      </Context.Provider>
    );
  };

  function eqlFn(prevValue: any, nextValue: any) {
    const r = prevValue === nextValue || shallowEqual(prevValue, nextValue);
    return r;
  }

  function useSelector<RR>(selector: (state: T) => RR) {
    const patchedSelector = useMemo(() => {
      let prevValue: any = null;
      return ([state]: T[]) => {
        const nextValue = selector(state);
        if (eqlFn(prevValue, nextValue)) {
          return prevValue;
        }
        return (prevValue = nextValue);
      };
    }, [selector]);
    return useContextSelector(Context, patchedSelector) as RR;
  }

  function useDispatch() {
    return useContextSelector(Context, (v) => v[1]) as <T extends keyof R>(
      _: Action<PayloadParams<R[T]>, T>
    ) => void;
  }

  return {
    Context,
    Provider,
    useSelector,
    useDispatch,
  };
}
